# -*- coding: utf-8 -*-
# Generated by Django 1.11.2.dev20170508150049 on 2017-05-09 05:45
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('raster', '0034_legendentryorder'),
    ]

    operations = [
        migrations.AddField(
            model_name='legendentry',
            name='code',
            field=models.CharField(blank=True, default='', max_length=100),
        ),
        migrations.AddField(
            model_name='legendentry',
            name='legend',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='raster.Legend'),
        ),
        migrations.AlterField(
            model_name='legend',
            name='entries',
            field=models.ManyToManyField(related_name='legentries', through='raster.LegendEntryOrder', to='raster.LegendEntry'),
        ),
    ]



from collections import Counter

import numpy

from django.contrib.gis.geos import MultiPolygon, Polygon
from django.core.exceptions import ObjectDoesNotExist
from raster.algebra.parser import FormulaParser, RasterAlgebraParser
from raster.exceptions import RasterAggregationException
from raster.models import Legend, RasterLayer
from raster.rasterize import rasterize
from raster.tiles.const import WEB_MERCATOR_SRID
from raster.tiles.lookup import get_raster_tile
from raster.tiles.utils import tile_index_range


class Aggregator(object):
    """
    Evaluate aggregation functions on raster layers.

    Functions such as value counts, max, min, mean or std are all values that
    depend on an entire layer, not single tiles. This class evaluates such
    functions on all tiles from a set of layers.
    """

    def __init__(self, layer_dict, formula, zoom=None, geom=None, acres=True,
                 grouping='auto', all_touched=True, memory_efficient=False, hist_range=None):
        # Set defining parameter for this aggregator
        self.layer_dict = layer_dict
        self.formula = formula
        self.geom = geom
        self.acres = acres
        self.rastgeom = None
        self.all_touched = all_touched
        self.memory_efficient = memory_efficient
        self.hist_range = hist_range

        # Get layers from input dict
        self.layers = RasterLayer.objects.filter(id__in=layer_dict.values())

        # Compute zoom if not provided
        if zoom is None:
            zoom = min(self.layers.values_list('metadata__max_zoom', flat=True))
        self.zoom = zoom

        # Compute tilerange for this area and the given zoom level
        if geom:
            # Transform geom to web mercator
            if geom.srid != WEB_MERCATOR_SRID:
                geom.transform(WEB_MERCATOR_SRID)

            # Clip against max extent for limiting nr of tiles.
            # This is important for requests on large areas for small rasters.
            max_extent = MultiPolygon([Polygon.from_bbox(lyr.extent()) for lyr in self.layers]).envelope
            max_extent = geom.intersection(max_extent)

            # Abort if there is no spatial overlay
            if max_extent.empty:
                self.tilerange = None
                return
            else:
                # Compute tile index range for geometry and given zoom level
                self.tilerange = tile_index_range(max_extent.extent, zoom)
        else:
            # Get index range set for the input layers
            index_ranges = [tile_index_range(lyr.extent(), zoom) for lyr in self.layers]

            # Compute intersection of index ranges
            self.tilerange = [
                max([dat[0] for dat in index_ranges]),
                max([dat[1] for dat in index_ranges]),
                min([dat[2] for dat in index_ranges]),
                min([dat[3] for dat in index_ranges])
            ]

        # Auto determine grouping based on input data
        if grouping == 'auto':
            all_discrete = all((lyr.datatype in (RasterLayer.CATEGORICAL, RasterLayer.MASK) for lyr in self.layers))
            grouping = 'discrete' if all_discrete else 'continuous'
        elif grouping in ('discrete', 'continuous'):
            pass
        else:
            try:
                legend_id = int(grouping)
                grouping = Legend.objects.get(id=legend_id)
            except ValueError:
                pass
            except ObjectDoesNotExist:
                raise RasterAggregationException(
                    'Invalid legend ID found in grouping value for valuecount.'
                )
        self.grouping = grouping

    def get_raster_tile(self, layerid, zoom, tilex, tiley):
        return get_raster_tile(layerid, zoom, tilex, tiley)

    def tiles(self):
        """
        Generator that yields an algebra-ready data dictionary for each tile in
        the aggregator's tile range.
        """
        # Check if any tiles have been matched
        if not self.tilerange:
            return

        algebra_parser = RasterAlgebraParser()

        for tilex in range(self.tilerange[0], self.tilerange[2] + 1):
            for tiley in range(self.tilerange[1], self.tilerange[3] + 1):

                # Prepare a data dictionary with named tiles for algebra evaluation
                data = {}
                for name, layerid in self.layer_dict.items():
                    tile = self.get_raster_tile(layerid, self.zoom, tilex, tiley)
                    if tile:
                        data[name] = tile
                    else:
                        break

                # Ignore this tile if it is missing in any of the input layers
                if len(data) < len(self.layer_dict):
                    continue

                # Compute raster algebra
                result = algebra_parser.evaluate_raster_algebra(data, self.formula)

                # Convert band data to masked array
                result_data = numpy.ma.masked_values(
                    result.bands[0].data(),
                    result.bands[0].nodata_value,
                )

                # Apply rasterized geometry as mask if clip geometry was provided
                if self.geom:
                    result_data = self.mask_by_geom(result, result_data)

                yield result_data.compressed()

    def _clear_stats(self):
        self._stats_t0 = 0
        self._stats_t1 = 0
        self._stats_t2 = 0
        self._stats_max_value = None
        self._stats_min_value = None

    def _push_stats(self, data):
        # Stop if entire data was masked
        if data.size == 0:
            return

        # Filter data by histogram range.
        if self.hist_range:
            stats_data = data[data >= self.hist_range[0]]
            stats_data = stats_data[stats_data <= self.hist_range[1]]
        else:
            stats_data = data

        # Compute incremental statistics
        self._stats_t0 += stats_data.size
        self._stats_t1 += numpy.sum(stats_data)
        self._stats_t2 += numpy.sum(numpy.square(stats_data))

        tile_max = numpy.max(stats_data)
        tile_min = numpy.min(stats_data)

        if self._stats_max_value is None:
            self._stats_max_value = tile_max
            self._stats_min_value = tile_min
        else:
            self._stats_max_value = max(tile_max, self._stats_max_value)
            self._stats_min_value = min(tile_min, self._stats_min_value)

    def statistics(self, reset=False):
        """
        Compute statistics for this aggregator. Returns (min, max, mean, std).
        The mean and std can be computed incrementally from the number of
        obeservations t0 = sum(x^0), the sum of values t1 = sum(x^1), and the
        sum of squares t2 = sum(x^2).
        """
        if reset or not hasattr(self, '_stats_max_value') or self._stats_max_value is None:
            self._clear_stats()
            # Only compute statistics if they have not been previously computed.
            for data in self.tiles():
                self._push_stats(data)

        if self._stats_t0 == 0:
            # If totals sum is zero, no data was available to comput statistics
            mean = None
            std = None
        else:
            # Compute mean and std from totals sums.
            mean = self._stats_t1 / self._stats_t0
            std = numpy.sqrt(self._stats_t0 * self._stats_t2 - self._stats_t1 * self._stats_t1) / self._stats_t0

        return (self._stats_min_value, self._stats_max_value, mean, std)

    def mask_by_geom(self, tile, data):
        # Rasterize the aggregation area to the result raster
        self.rastgeom = rasterize(self.geom, tile, all_touched=self.all_touched)

        # Get boolean mask based on rasterized geom
        rastgeom_mask = self.rastgeom.bands[0].data() != 1

        # Apply geometry mask to result data
        data.mask = data.mask | rastgeom_mask

        return data

    def value_count(self):
        """
        Compute aggregate statistics for a layers dictionary, potentially for
        an algebra expression and clipped by a geometry.

        The grouping parameter specifies how to group the pixel values for the
        aggregation count.

        Allowed are the following options:

        * 'auto' (the default). The output will be grouped by unique values if all
          input rasters are discrete, otherwise a numpy histogram is used.
        * 'discrete' groups the data will be grouped by unique values
        * 'continuous' groups the data in a numpy histogram
        * If an integer value is passed to the argument, it is interpreted as a
          legend_id. The data will be grouped using the legend expressions. For
          For instance, use grouping=23 for grouping the output with legend 23.
        """
        # Instantiate counter dict for value counts.
        results = Counter({})
        self._clear_stats()

        if self.memory_efficient:
            # Loop through tiles individually.
            all_result_data = self.tiles()
        else:
            # Combine all tiles into one big array.
            all_result_data = [tile for tile in self.tiles()]
            if len(all_result_data):
                all_result_data = (
                    numpy.concatenate(all_result_data),
                )

        for result_data in all_result_data:

            if self.grouping == 'discrete':
                # Compute unique counts for discrete input data
                unique_counts = numpy.unique(result_data, return_counts=True)
                # Add counts to results
                values = dict(zip(unique_counts[0], unique_counts[1]))

            elif self.grouping == 'continuous':
                if self.memory_efficient and not self.hist_range:
                    raise RasterAggregationException(
                        'Secify a histogram range for memory efficient continuous aggregation.'
                    )

                # Handle continuous case - compute histogram on masked data
                counts, bins = numpy.histogram(result_data, range=self.hist_range)

                # Create dictionary with bins as keys and histogram counts as values
                values = {}
                for i in range(len(bins) - 1):
                    values[(bins[i], bins[i + 1])] = counts[i]

            else:
                # If input is not a legend, interpret input as legend json data
                if not isinstance(self.grouping, Legend):
                    self.grouping = Legend(json=self.grouping)

                # Try getting a colormap from the input
                try:
                    colormap = self.grouping.colormap
                except:
                    raise RasterAggregationException(
                        'Invalid grouping value found for valuecount.'
                    )

                # Use colormap to compute value counts
                formula_parser = FormulaParser()
                values = {}
                for key, color in colormap.items():
                    try:
                        # Try to use the key as number directly
                        selector = result_data == float(key)
                    except ValueError:
                        # Otherwise use it as numpy expression directly
                        selector = formula_parser.evaluate({'x': result_data}, key)
                    values[key] = numpy.sum(selector)

            # Add counts to results.
            results.update(Counter(values))
            # Push statistics.
            self._push_stats(result_data)

        # Transform pixel count to acres if requested
        scaling_factor = 1
        if self.acres and self.rastgeom and len(results):
            scaling_factor = abs(self.rastgeom.scale.x * self.rastgeom.scale.y) * 0.000247105381

        results = {
            str(int(k) if type(k) == numpy.float64 and int(k) == k else k):
            v * scaling_factor for k, v in results.items()
        }

        return results



# -*- coding: utf-8 -*-
import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('raster', '0023_remove_rasterlayer_parse_log'),
    ]

    operations = [
        migrations.CreateModel(
            name='RasterLayerBandMetadata',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('band', models.PositiveIntegerField()),
                ('nodata_value', models.FloatField(null=True)),
                ('max', models.FloatField()),
                ('min', models.FloatField()),
                ('hist_values', django.contrib.postgres.fields.ArrayField(base_field=models.FloatField(), size=100)),
                ('hist_bins', django.contrib.postgres.fields.ArrayField(base_field=models.FloatField(), size=101)),
            ],
        ),
        migrations.AlterField(
            model_name='rasterlayer',
            name='nodata',
            field=models.CharField(max_length=100, null=True, blank=True),
        ),
        migrations.AddField(
            model_name='rasterlayerbandmetadata',
            name='rasterlayer',
            field=models.ForeignKey(to='raster.RasterLayer', on_delete=models.CASCADE),
        ),
    ]



# -*- coding: utf-8 -*-
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('raster', '0019_remove_rasterlayer_srid'),
    ]

    operations = [
        migrations.AddField(
            model_name='rasterlayermetadata',
            name='max_zoom',
            field=models.PositiveSmallIntegerField(null=True, blank=True),
        ),
        migrations.AlterField(
            model_name='rasterlayermetadata',
            name='rasterlayer',
            field=models.OneToOneField(related_name='metadata', to='raster.RasterLayer', on_delete=models.CASCADE),
        ),
    ]



# -*- coding: utf-8 -*-
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('raster', '0028_auto_20160104_0533'),
    ]

    operations = [
        migrations.AddField(
            model_name='rasterlayerbandmetadata',
            name='mean',
            field=models.FloatField(null=True),
        ),
        migrations.AddField(
            model_name='rasterlayerbandmetadata',
            name='std',
            field=models.FloatField(null=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.3.dev20160204092059 on 2016-02-18 08:32
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('raster', '0030_auto_20160111_0813'),
    ]

    operations = [
        migrations.AddField(
            model_name='rasterlayer',
            name='max_zoom',
            field=models.IntegerField(blank=True, help_text=b'Leave blank to automatically determine the max zoom levelfrom the raster scale. Otherwise the raster parsed up tothe zoom level specified here.', null=True),
        ),
        migrations.AddField(
            model_name='rasterlayer',
            name='srid',
            field=models.IntegerField(blank=True, help_text=b'Leave blank to use the internal raster srid. If a srid isspecified here, it will be used for all calculations.', null=True),
        ),
    ]



# -*- coding: utf-8 -*-
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('raster', '0026_auto_20151120_1334'),
    ]

    operations = [
        migrations.CreateModel(
            name='RasterLayerReprojected',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('rasterfile', models.FileField(null=True, upload_to=b'rasters/reprojected', blank=True)),
                ('rasterlayer', models.OneToOneField(related_name='reprojected', to='raster.RasterLayer', on_delete=models.CASCADE)),
            ],
        ),
    ]



# -*- coding: utf-8 -*-
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('raster', '0004_rasterlayermetadata'),
    ]

    operations = [
        migrations.RunSQL(
            "DROP INDEX IF EXISTS raster_rastertile_rast_st_convexhull_idx; CREATE INDEX raster_rastertile_rast_st_convexhull_idx\
             ON raster_rastertile USING gist( ST_ConvexHull(rast) )",
            "DROP INDEX IF EXISTS raster_rastertile_rast_st_convexhull_idx"
        )
    ]



# -*- coding: utf-8 -*-
import colorful.fields

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('raster', '0007_auto_20141017_0240'),
    ]

    operations = [
        migrations.CreateModel(
            name='Legend',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('title', models.CharField(max_length=200)),
                ('description', models.TextField(null=True, blank=True)),
                ('json', models.TextField(null=True, blank=True)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='LegendEntry',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('expression', models.CharField(max_length=500)),
                ('color', colorful.fields.RGBColorField()),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='LegendSemantics',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=50)),
                ('description', models.TextField(null=True, blank=True)),
                ('keyword', models.TextField(max_length=100, null=True, blank=True)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name='legendentry',
            name='semantics',
            field=models.ForeignKey(to='raster.LegendSemantics', on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='legend',
            name='entries',
            field=models.ManyToManyField(to='raster.LegendEntry'),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='rasterlayer',
            name='legend',
            field=models.ForeignKey(blank=True, to='raster.Legend', null=True, on_delete=models.CASCADE),
            preserve_default=True,
        ),
    ]



import json
import math
from operator import itemgetter

import numpy
from colorful.fields import RGBColorField

from django.contrib.gis.db import models
from django.contrib.gis.gdal import Envelope, OGRGeometry, SpatialReference
from django.contrib.postgres.fields import ArrayField
from django.db.models import Max, Min
from django.db.models.signals import post_delete, post_save, pre_save
from django.dispatch import receiver
from raster.mixins import ValueCountMixin
from raster.tiles.const import WEB_MERCATOR_SRID
from raster.utils import hex_to_rgba


class LegendSemantics(models.Model):
    """
    Labels for pixel types (urban, forest, warm, cold, etc)
    """
    name = models.CharField(max_length=50)
    description = models.TextField(null=True, blank=True)
    keyword = models.TextField(null=True, blank=True, max_length=100)

    def __str__(self):
        return self.name


class LegendEntry(models.Model):
    """
    One row in a Legend.
    """
    legend = models.ForeignKey('Legend', null=True, on_delete=models.CASCADE)
    semantics = models.ForeignKey(LegendSemantics, on_delete=models.CASCADE)
    expression = models.CharField(max_length=500,
        help_text='Use a number or a valid numpy logical expression where x is the'
                  'pixel value. For instance: "(-3.0 < x) & (x <= 1)" or "x <= 1".')
    color = RGBColorField()
    code = models.CharField(max_length=100, blank=True, default='')

    def __str__(self):
        return '{}, {}, {}'.format(self.semantics.name, self.expression, self.color)


class Legend(models.Model):
    """
    Legend object for Rasters.
    """
    title = models.CharField(max_length=200)
    description = models.TextField(null=True, blank=True)
    json = models.TextField(null=True, blank=True)
    modified = models.DateTimeField(auto_now=True)

    def __str__(self):
        return self.title

    def update_json(self):
        data = []
        for entry in self.legendentry_set.all():
            data.append({
                'name': entry.semantics.name,
                'expression': entry.expression,
                'code': entry.code,
                'color': entry.color,
            })
        # Sort data by name and code.
        data = sorted(sorted(data, key=itemgetter('name')), key=itemgetter('code'))
        self.json = json.dumps(data)

    @property
    def colormap(self):
        legend = json.loads(self.json)
        cmap = {}
        for leg in legend:
            cmap[leg['expression']] = hex_to_rgba(leg['color'])
        return cmap

    def save(self, *args, **kwargs):
        if self.id:
            self.update_json()
        super(Legend, self).save(*args, **kwargs)


@receiver(post_save, sender=LegendEntry)
@receiver(post_delete, sender=LegendEntry)
def legend_entries_changed(sender, instance, **kwargs):
    """
    Updates style json upon adding or removing legend entries.
    """
    try:
        instance.legend.update_json()
        instance.legend.save()
    except Legend.DoesNotExist:
        pass


@receiver(post_save, sender=LegendSemantics)
def update_dependent_legends_on_semantics_change(sender, instance, **kwargs):
    """
    Updates dependent Legends on a change in Semantics.
    """
    legend_ids = LegendEntry.objects.filter(semantics_id=instance.id).values_list('legend_id', flat=True)
    for legend in Legend.objects.filter(id__in=legend_ids):
        legend.update_json()
        legend.save()


class RasterLayer(models.Model, ValueCountMixin):
    """
    Source data model for raster layers
    """
    CONTINUOUS = 'co'
    CATEGORICAL = 'ca'
    MASK = 'ma'
    RANK_ORDERED = 'ro'

    DATATYPES = (
        (CONTINUOUS, 'Continuous'),
        (CATEGORICAL, 'Categorical'),
        (MASK, 'Mask'),
        (RANK_ORDERED, 'Rank Ordered'),
    )

    name = models.CharField(max_length=100, blank=True, null=True)
    description = models.TextField(blank=True, null=True)
    datatype = models.CharField(max_length=2, choices=DATATYPES, default='co')
    rasterfile = models.FileField(upload_to='rasters', null=True, blank=True)
    source_url = models.CharField(default='', blank=True, max_length=2500,
        help_text='External url to get the raster file from. If a value is set,'
                  'the rasterfile field will be ignored.')
    nodata = models.CharField(max_length=100, null=True, blank=True,
        help_text='Leave blank to keep the internal band nodata values. If a nodata '
                  'value is specified here, it will be used for all bands of this raster.')
    srid = models.IntegerField(null=True, blank=True,
        help_text='Leave blank to use the internal raster srid. If a srid is '
                  'specified here, it will be used for all calculations.')
    max_zoom = models.IntegerField(null=True, blank=True,
        help_text='Leave blank to automatically determine the max zoom level '
                  'from the raster scale. Otherwise the raster parsed up to '
                  'the zoom level specified here.')
    build_pyramid = models.BooleanField(default=True,
        help_text='Should the tile pyramid be built? If unchecked, tiles will '
                  'only be generated at the max zoom level.')
    next_higher = models.BooleanField(default=True,
        help_text='Compared to the scale of the rasterlayer, use the next-higher '
                  'zoomlevel as max zoom? If unchecked, the next-lower zoom level '
                  'is used. This flag is ignored if the max_zoom is manually '
                  'specified.')
    store_reprojected = models.BooleanField(default=True,
        help_text='Should the reprojected raster be stored? If unchecked, the '
                  'reprojected version of the raster is not stored.')
    legend = models.ForeignKey(Legend, blank=True, null=True, on_delete=models.CASCADE)
    modified = models.DateTimeField(auto_now=True)

    def __str__(self):
        return '{} {} (type: {})'.format(self.id, self.name, self.datatype)

    @property
    def discrete(self):
        """
        Returns true for discrete rasters.
        """
        return self.datatype in (self.CATEGORICAL, self.MASK, self.RANK_ORDERED)

    _bbox = None
    _bbox_srid = None

    def extent(self, srid=WEB_MERCATOR_SRID):
        """
        Returns bbox for layer.
        """
        if not self._bbox or self._bbox_srid != srid:
            # Get bbox for raster in original coordinates
            meta = self.metadata
            xmin = meta.uperleftx
            ymax = meta.uperlefty
            xmax = xmin + meta.width * meta.scalex
            ymin = ymax + meta.height * meta.scaley

            # Create Polygon box
            geom = OGRGeometry(Envelope((xmin, ymin, xmax, ymax)).wkt)

            # Set original srs
            if meta.srs_wkt:
                geom.srs = SpatialReference(meta.srs_wkt)
            else:
                geom.srid = meta.srid

            # Transform to requested srid
            geom.transform(srid)

            # Calculate value range for bbox
            coords = geom.coords[0]
            xvals = [x[0] for x in coords]
            yvals = [x[1] for x in coords]

            # Set bbox
            self._bbox = (min(xvals), min(yvals), max(xvals), max(yvals))
            self._bbox_srid = srid
        return self._bbox

    def index_range(self, zoom):
        """
        Compute the index range for this rasterlayer at a given zoom leve.
        """
        return self.rastertile_set.filter(tilez=zoom).aggregate(
            Min('tilex'), Max('tilex'), Min('tiley'), Max('tiley')
        )


@receiver(pre_save, sender=RasterLayer)
def reset_parse_log_if_data_changed(sender, instance, **kwargs):
    try:
        obj = RasterLayer.objects.get(pk=instance.pk)
    except RasterLayer.DoesNotExist:
        pass
    else:
        # If filename or nodata value has changed, clear parse status to
        # trigger re-parsing. Also remove the reprojected copy of the previous
        # file if it exists.
        if (obj.rasterfile.name != instance.rasterfile.name
                or obj.nodata != instance.nodata
                or obj.max_zoom != instance.max_zoom
                or obj.srid != instance.srid
                or obj.source_url != instance.source_url
                or obj.build_pyramid != instance.build_pyramid
                or obj.next_higher != instance.next_higher):
            if hasattr(instance, 'reprojected'):
                instance.reprojected.delete()
            instance.parsestatus.reset()


@receiver(post_save, sender=RasterLayer)
def parse_raster_layer_if_status_is_unparsed(sender, instance, created, **kwargs):
    from raster.tasks import parse
    RasterLayerMetadata.objects.get_or_create(rasterlayer=instance)
    status, created = RasterLayerParseStatus.objects.get_or_create(rasterlayer=instance)
    if (instance.rasterfile.name or instance.source_url) and status.status == status.UNPARSED:
        parse(instance.id)


class RasterLayerReprojected(models.Model):
    """
    Stores reprojected version of raster.
    """
    rasterlayer = models.OneToOneField(RasterLayer, related_name='reprojected', on_delete=models.CASCADE)
    rasterfile = models.FileField(upload_to='rasters/reprojected', null=True, blank=True)

    def __str__(self):
        return self.rasterlayer.name


class RasterLayerMetadata(models.Model):
    """
    Stores meta data for a raster layer
    """
    rasterlayer = models.OneToOneField(RasterLayer, related_name='metadata', on_delete=models.CASCADE)
    uperleftx = models.FloatField(null=True, blank=True)
    uperlefty = models.FloatField(null=True, blank=True)
    width = models.IntegerField(null=True, blank=True)
    height = models.IntegerField(null=True, blank=True)
    scalex = models.FloatField(null=True, blank=True)
    scaley = models.FloatField(null=True, blank=True)
    skewx = models.FloatField(null=True, blank=True)
    skewy = models.FloatField(null=True, blank=True)
    numbands = models.IntegerField(null=True, blank=True)
    srs_wkt = models.TextField(null=True, blank=True)
    srid = models.PositiveSmallIntegerField(null=True, blank=True)
    max_zoom = models.PositiveSmallIntegerField(null=True, blank=True)

    def __str__(self):
        return self.rasterlayer.name


class RasterLayerParseStatus(models.Model):
    """
    Tracks the parsing status of the raster layer.
    """
    UNPARSED = 0
    DOWNLOADING_FILE = 1
    REPROJECTING_RASTER = 2
    CREATING_TILES = 3
    DROPPING_EMPTY_TILES = 4
    FINISHED = 5
    FAILED = 6

    STATUS_CHOICES = (
        (UNPARSED, 'Layer not yet parsed'),
        (DOWNLOADING_FILE, 'Downloading file'),
        (REPROJECTING_RASTER, 'Reprojecting'),
        (CREATING_TILES, 'Creating tiles'),
        (DROPPING_EMPTY_TILES, 'Dropping empty tiles'),
        (FINISHED, 'Finished parsing'),
        (FAILED, 'Failed parsing'),
    )
    rasterlayer = models.OneToOneField(RasterLayer, related_name='parsestatus', on_delete=models.CASCADE)
    status = models.IntegerField(choices=STATUS_CHOICES, default=UNPARSED)
    log = models.TextField(default='', editable=False)
    tile_levels = ArrayField(models.PositiveIntegerField(), default=list)

    def __str__(self):
        return '{0} - {1}'.format(self.rasterlayer.name, self.get_status_display())

    def reset(self):
        self.tile_levels = []
        self.log = ''
        self.status = self.UNPARSED
        self.save()


class RasterLayerBandMetadata(models.Model):

    HISTOGRAM_BINS = 100

    rasterlayer = models.ForeignKey(RasterLayer, on_delete=models.CASCADE)
    band = models.PositiveIntegerField()
    nodata_value = models.FloatField(null=True)
    min = models.FloatField()
    max = models.FloatField()
    mean = models.FloatField(null=True)
    std = models.FloatField(null=True)
    hist_values = ArrayField(models.BigIntegerField(), size=HISTOGRAM_BINS)
    hist_bins = ArrayField(models.FloatField(), size=HISTOGRAM_BINS + 1)

    class Meta:
        unique_together = ('rasterlayer', 'band')

    def __str__(self):
        return '{} - Min {} - Max {}'.format(self.rasterlayer.name, self.min, self.max)

    def save(self, *args, **kwargs):
        if not self.pk:
            # Construct empty histogram
            hist = numpy.histogram(
                [],
                range=(math.floor(self.min), math.ceil(self.max)),
                bins=self.HISTOGRAM_BINS
            )
            # Set empty histogram values
            self.hist_values = hist[0].tolist()
            self.hist_bins = hist[1].tolist()

        super(RasterLayerBandMetadata, self).save(*args, **kwargs)

    def statistics(self):
        return (self.min, self.max, self.mean, self.std)


class RasterTile(models.Model):
    """
    Store individual tiles of a raster data source layer.
    """
    ZOOMLEVELS = (
        (1, 1), (2, 2), (3, 3), (4, 4), (5, 5), (6, 6), (7, 7),
        (8, 8), (9, 9), (10, 10), (11, 11), (12, 12), (13, 13),
        (14, 14), (15, 15), (16, 16), (17, 17), (18, 18)
    )
    rid = models.AutoField(primary_key=True)
    rast = models.RasterField(null=True, blank=True, srid=WEB_MERCATOR_SRID)
    rasterlayer = models.ForeignKey(RasterLayer, null=True, blank=True, db_index=True, on_delete=models.CASCADE)
    tilex = models.IntegerField(db_index=True, null=True)
    tiley = models.IntegerField(db_index=True, null=True)
    tilez = models.IntegerField(db_index=True, null=True, choices=ZOOMLEVELS)

    def __str__(self):
        return '{} {}'.format(self.rid, self.rasterlayer.name)



# -*- coding: utf-8 -*-
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('raster', '0008_auto_20150219_0413'),
    ]

    operations = [
        migrations.AddField(
            model_name='rasterlayer',
            name='max_zoom',
            field=models.IntegerField(null=True, blank=True),
            preserve_default=True,
        ),
    ]



"""
Everything required to create TMS tiles.
"""
from django.conf import settings
from raster.tiles.const import (
    GLOBAL_MAX_ZOOM_LEVEL, QUADRANT_SIZE, WEB_MERCATOR_TILESHIFT, WEB_MERCATOR_TILESIZE, WEB_MERCATOR_WORLDSIZE
)


def tile_index_range(bbox, z, tolerance=0):
    """
    Calculate the index range for a given bounding box and zoomlevel. The bbox
    coordinages are assumed to be in Web Mercator.

    The strict option can be used to force only strict overlaps, based on a
    tolerance.
    """
    # Calculate tile size for given zoom level.
    zscale = WEB_MERCATOR_WORLDSIZE / 2 ** z

    # Calculate overlaying tile indices.
    result_float = [
        (bbox[0] + WEB_MERCATOR_TILESHIFT) / zscale,
        (WEB_MERCATOR_TILESHIFT - bbox[3]) / zscale,
        (bbox[2] + WEB_MERCATOR_TILESHIFT) / zscale,
        (WEB_MERCATOR_TILESHIFT - bbox[1]) / zscale,
    ]
    # Use integer floor as index. This ensures overlap, since
    # the idex values are counted from the upper left corner.
    result = [None] * 4

    for i in range(4):
        # If the index range is a close call, make sure that only
        # strictly overlapping indices are included.
        if abs(round(result_float[i]) - result_float[i]) < tolerance:
            result[i] = round(result_float[i])
            # For the max range values, reduce so that the edge tile is not
            # included.
            if i > 1:
                result[i] -= 1
        else:
            result[i] = int(result_float[i])

    return result


def tile_bounds(x, y, z):
    """
    Calculate the bounding box of a specific tile.
    """
    zscale = WEB_MERCATOR_WORLDSIZE / 2 ** z

    xmin = x * zscale - WEB_MERCATOR_TILESHIFT
    xmax = (x + 1) * zscale - WEB_MERCATOR_TILESHIFT
    ymin = WEB_MERCATOR_TILESHIFT - (y + 1) * zscale
    ymax = WEB_MERCATOR_TILESHIFT - y * zscale

    return [xmin, ymin, xmax, ymax]


def tile_scale(z):
    """
    Calculate tile pixel size scale for given zoom level.
    """
    TILESIZE = int(getattr(settings, 'RASTER_TILESIZE', WEB_MERCATOR_TILESIZE))
    return WEB_MERCATOR_WORLDSIZE / 2.0 ** z / TILESIZE


def closest_zoomlevel(scale, next_higher=True):
    """
    Calculate the zoom level index z that is closest to the given scale.
    The input scale needs to be provided in meters per pixel. It is then
    compared to a list of pixel sizes for all TMS zoom levels.
    """
    TILESIZE = int(getattr(settings, 'RASTER_TILESIZE', WEB_MERCATOR_TILESIZE))
    # Calculate all pixelsizes for the TMS zoom levels
    tms_pixelsizes = [WEB_MERCATOR_WORLDSIZE / (2.0 ** (i + 1) * TILESIZE) for i in range(GLOBAL_MAX_ZOOM_LEVEL)]

    # If the pixelsize is smaller than all tms sizes, default to max level
    zoomlevel = GLOBAL_MAX_ZOOM_LEVEL

    # Find zoomlevel (next-upper) for the input pixel size
    for i in range(0, GLOBAL_MAX_ZOOM_LEVEL):
        if scale - tms_pixelsizes[i] >= 0:
            zoomlevel = i
            break

    # If nextdown setting is true, adjust level
    if next_higher:
        zoomlevel += 1

    return zoomlevel


def quadrants(bbox, z):
    """
    Create an array of bounding boxes, representing a set of sub-regions
    defined as tile index ranges that cover the input bounding box. This
    is used to create tiles on quadrants instead of the entire file at once.
    """
    indexrange = tile_index_range(bbox, z)
    quadrant_list = []

    for tilex in range(indexrange[0], indexrange[2] + 1, QUADRANT_SIZE):
        for tiley in range(indexrange[1], indexrange[3] + 1, QUADRANT_SIZE):
            quadrant_list.append((
                tilex,
                tiley,
                min(tilex + QUADRANT_SIZE - 1, indexrange[2]),
                min(tiley + QUADRANT_SIZE - 1, indexrange[3]),
            ))

    return quadrant_list



# -*- coding: utf-8 -*-
import django.contrib.gis.db.models.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('raster', '0010_auto_20150429_1207'),
    ]

    operations = [
        migrations.AlterField(
            model_name='rastertile',
            name='rast',
            field=django.contrib.gis.db.models.fields.RasterField(srid=3857, null=True, blank=True),
        ),
    ]



from django.conf.urls import url
from raster.views import AlgebraView, ExportView, LegendView

urlpatterns = [

    # Normal raster tiles endpoint
    url(
        r'^tiles/(?P<layer>[^/]+)/(?P<z>[0-9]+)/(?P<x>[0-9]+)/(?P<y>[0-9]+).(?P<frmt>png|jpg|tif)$',
        AlgebraView.as_view(),
        name='tms',
    ),

    # Raster algebra endpoint
    url(
        r'^algebra/(?P<z>[0-9]+)/(?P<x>[0-9]+)/(?P<y>[0-9]+).(?P<frmt>jpg|png|tif)$',
        AlgebraView.as_view(),
        name='algebra',
    ),

    # Pixel value endpoint
    url(
        r'^pixel/(?P<xcoord>-?\d+(?:\.\d+)?)/(?P<ycoord>-?\d+(?:\.\d+)?)$',
        AlgebraView.as_view(),
        name='pixel',
    ),

    # Raster legend endpoint
    url(
        r'^legend(?:/(?P<legend_id>[^/]+))?/$',
        LegendView.as_view(),
        name='legend',
    ),

    # Exporter endpoint
    url(
        r'^export$',
        ExportView.as_view(),
        name='export',
    ),
]



# -*- coding: utf-8 -*-
from django.contrib.gis.db import models
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='RasterLayer',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=100, null=True, blank=True)),
                ('description', models.TextField(null=True, blank=True)),
                ('datatype', models.CharField(default=b'co', max_length=2, choices=[(b'co', b'Continuous'), (b'ca', b'Categorical'), (b'ma', b'Mask'), (b'ro', b'Rank Ordered')])),
                ('rasterfile', models.FileField(upload_to=b'rasters')),
                ('srid', models.CharField(default=b'3086', max_length=10)),
                ('nodata', models.CharField(default=b'-9999', max_length=100)),
                ('parse_log', models.TextField(default=b'', null=True, editable=False, blank=True)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='RasterTile',
            fields=[
                ('rid', models.AutoField(serialize=False, primary_key=True)),
                ('rast', models.RasterField(null=True, blank=True)),
                ('filename', models.TextField(null=True, blank=True)),
                ('rasterlayer', models.ForeignKey(blank=True, to='raster.RasterLayer', null=True, on_delete=models.CASCADE)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
    ]



"""
Settings for django-raster tests.
"""
import os

SECRET_KEY = 'testkey'

INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.postgres',
    'django.contrib.gis',

    'raster',
)

DATABASES = {
    'default': {
        'ENGINE': 'django.contrib.gis.db.backends.postgis',
        'USER': os.environ.get('DB_USER', 'postgres'),
        'HOST': os.environ.get('DB_HOST', 'localhost'),
        'NAME': os.environ.get('DB_NAME', 'postgres'),
        'PASSWORD': os.environ.get('DB_PASSWORD', ''),
        'PORT': os.environ.get('DB_PORT', '5432')
    }
}

MIDDLEWARE = [
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
]

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

ROOT_URLCONF = 'raster.urls'

RASTER_USE_CELERY = True

CELERY_TASK_ALWAYS_EAGER = True
CELERY_TASK_EAGER_PROPAGATES = True



from django.contrib.gis.geos import GEOSGeometry
from django.db import connection
from raster.tiles.const import WEB_MERCATOR_SRID

CLIPPED_VALUE_COUNT_SQL = """
WITH tiles_for_agg AS (
    SELECT ST_ValueCount(ST_Clip(ST_Transform(rast, {geom_srid}), ST_GeomFromEWKT('{geom_ewkt}'))) AS vcresult
    FROM raster_rastertile
    WHERE ST_Intersects(rast, ST_Transform(ST_GeomFromEWKT('{geom_ewkt}'), {rast_srid}))
    AND rasterlayer_id = {rasterlayer_id}
    AND tilez = {zoom}
)
SELECT (vcresult).value, SUM((vcresult).count) AS count
FROM tiles_for_agg
GROUP BY (vcresult).value
"""

GLOBAL_VALUE_COUNT_SQL = """
WITH tiles_for_agg AS (
    SELECT ST_ValueCount(rast) AS vcresult
    FROM raster_rastertile
    WHERE rasterlayer_id = {rasterlayer_id}
    AND tilez = {zoom}
)
SELECT (vcresult).value, SUM((vcresult).count) AS count
FROM tiles_for_agg
GROUP BY (vcresult).value
"""

MINSIZE_SQL = """
SELECT
    ST_ScaleX(ST_Transform(rast, {srid})) AS scalex,
    ST_ScaleY(ST_Transform(rast, {srid})) AS scaley
FROM raster_rastertile
WHERE rasterlayer_id = {rasterlayer_id}
AND tilez = {zoom}
LIMIT 1
"""

MAX_ZOOM_SQL = """
SELECT MAX(tilez)
FROM raster_rastertile
WHERE rasterlayer_id={rasterlayer_id}
"""


class ValueCountMixin(object):
    """
    Value count methods for Raster Layers.
    """
    def db_value_count(self, geom=None, area=False, zoom=None):
        """
        Compute value count in database.
        """
        if not zoom:
            zoom = self._max_zoom

        # Check that raster is categorical or mask
        if self.datatype not in ['ca', 'ma']:
            raise TypeError(
                'Wrong rastertype, value counts can only be '
                'calculated for categorical or mask raster types'
            )

        if geom:
            # Make sure geometry is GEOS Geom
            geom = GEOSGeometry(geom)

            sql = CLIPPED_VALUE_COUNT_SQL.format(
                geom_ewkt=geom.ewkt,
                geom_srid=geom.srid,
                rast_srid=WEB_MERCATOR_SRID,
                rasterlayer_id=self.id,
                zoom=zoom
            )
        else:
            sql = GLOBAL_VALUE_COUNT_SQL.format(
                rasterlayer_id=self.id,
                zoom=zoom
            )

        cursor = connection.cursor()
        cursor.execute(sql)

        # Convert value count to areas if requested
        if area:
            scalex, scaley = self._min_pixelsize(geom.srid)
            return {int(row[0]): int(row[1]) * scalex * scaley for row in cursor.fetchall()}
        else:
            return {int(row[0]): int(row[1]) for row in cursor.fetchall()}

    _maxz = None

    @property
    def _max_zoom(self):
        """
        Get max zoom for this layer.
        """
        if not self._maxz:
            cursor = connection.cursor()
            cursor.execute(MAX_ZOOM_SQL.format(rasterlayer_id=self.id))
            self._maxz = cursor.fetchone()[0]
        return self._maxz

    def pixelsize(self, srid=WEB_MERCATOR_SRID, zoom=None):
        """
        Compute size of a pixel for a given srid and zoomlevel.
        """
        if not zoom:
            zoom = self._max_zoom

        sql = MINSIZE_SQL.format(
            srid=srid,
            rasterlayer_id=self.id,
            zoom=zoom
        )

        cursor = connection.cursor()
        cursor.execute(sql)
        res = cursor.fetchone()
        self._minsize = (abs(res[0]), abs(res[1]))
        self._minsize_srid = srid

        return self._minsize

    def value_count(self, geom=None, area=False, zoom=None):
        """
        Compute value counts or histograms for rasterlayers within a geometry.
        """
        from raster.valuecount import Aggregator
        agg = Aggregator(layer_dict={'a': self.id}, formula='a', zoom=zoom, geom=geom, acres=area)
        return agg.value_count()



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-26 01:52
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('raster', '0031_auto_20160218_0832'),
    ]

    operations = [
        migrations.AlterField(
            model_name='legendentry',
            name='expression',
            field=models.CharField(help_text='Use a number or a valid numpy logical expression where x is thepixel value. For instance: "(-3.0 < x) & (x <= 1)" or "x <= 1".', max_length=500),
        ),
        migrations.AlterField(
            model_name='rasterlayer',
            name='datatype',
            field=models.CharField(choices=[('co', 'Continuous'), ('ca', 'Categorical'), ('ma', 'Mask'), ('ro', 'Rank Ordered')], default='co', max_length=2),
        ),
        migrations.AlterField(
            model_name='rasterlayer',
            name='max_zoom',
            field=models.IntegerField(blank=True, help_text='Leave blank to automatically determine the max zoom level from the raster scale. Otherwise the raster parsed up to the zoom level specified here.', null=True),
        ),
        migrations.AlterField(
            model_name='rasterlayer',
            name='nodata',
            field=models.CharField(blank=True, help_text='Leave blank to keep the internal band nodata values. If a nodata value is specified here, it will be used for all bands of this raster.', max_length=100, null=True),
        ),
        migrations.AlterField(
            model_name='rasterlayer',
            name='rasterfile',
            field=models.FileField(blank=True, null=True, upload_to='rasters'),
        ),
        migrations.AlterField(
            model_name='rasterlayer',
            name='srid',
            field=models.IntegerField(blank=True, help_text='Leave blank to use the internal raster srid. If a srid is specified here, it will be used for all calculations.', null=True),
        ),
        migrations.AlterField(
            model_name='rasterlayerparsestatus',
            name='log',
            field=models.TextField(default='', editable=False),
        ),
        migrations.AlterField(
            model_name='rasterlayerparsestatus',
            name='status',
            field=models.IntegerField(choices=[(0, 'Layer not yet parsed'), (1, 'Downloading file'), (2, 'Reprojecting'), (3, 'Creating tiles'), (4, 'Dropping empty tiles'), (5, 'Finished parsing'), (6, 'Failed parsing')], default=0),
        ),
        migrations.AlterField(
            model_name='rasterlayerreprojected',
            name='rasterfile',
            field=models.FileField(blank=True, null=True, upload_to='rasters/reprojected'),
        ),
    ]



# -*- coding: utf-8 -*-
import datetime

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('raster', '0014_remove_rastertile_is_base'),
    ]

    operations = [
        migrations.AddField(
            model_name='legend',
            name='modified',
            field=models.DateTimeField(default=datetime.datetime(2015, 8, 19, 3, 20, 29, 439201), auto_now=True),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='rasterlayer',
            name='modified',
            field=models.DateTimeField(default=datetime.datetime(2015, 8, 19, 3, 20, 37, 663400), auto_now=True),
            preserve_default=False,
        ),
    ]



# -*- coding: utf-8 -*-
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('raster', '0012_auto_20150616_0538'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='rasterlayer',
            name='max_zoom',
        ),
    ]



from PIL import ImageEnhance

IMG_ENHANCEMENTS = {
    'enhance_color': ImageEnhance.Color,
    'enhance_contrast': ImageEnhance.Contrast,
    'enhance_brightness': ImageEnhance.Brightness,
    'enhance_sharpness': ImageEnhance.Sharpness,
}
IMG_FORMATS = {'png': ('PNG', 'image/png'), 'jpg': ('JPEG', 'image/jpeg'), 'tif': ('TIFF', 'image/tiff')}
EXPORT_MAX_PIXELS = 10000 * 10000
MAX_EXPORT_NAME_LENGTH = 100
README_TEMPLATE = """Django Raster Algebra Export
============================
{description}
Input layers
------------
{layers}
Raster algebra formula
----------------------
{formula}

Geographic parameters
---------------------
Zoom level: {zoom}
Tile index range x: {xindexrange}
Tile index range y: {yindexrange}
Bounding-box: {bbox}

Source
------
Files auto-generated on {datetime} from the following url:

{url}
"""



# -*- coding: utf-8 -*-
from django.db import migrations


def move_srid_from_layer_to_metadata_forward(apps, schema_editor):
    """
    Copy srid from rasterlayers to metadata as a default value. This value will
    be overwritten automatically when parsing the layer.
    """
    RasterLayer = apps.get_model("raster", "RasterLayer")
    RasterLayerMetadata = apps.get_model("raster", "RasterLayerMetadata")
    for lyr in RasterLayer.objects.all():
        meta, created = RasterLayerMetadata.objects.get_or_create(rasterlayer=lyr)
        meta.srid = lyr.srid
        meta.save()


def move_srid_from_layer_to_metadata_backward(apps, schema_editor):
    """
    Copy the srids back to the raster layers.
    """
    RasterLayer = apps.get_model("raster", "RasterLayer")
    for lyr in RasterLayer.objects.all():
        if hasattr(lyr, 'rasterlayermetadata'):
            lyr.srid = lyr.rasterlayermetadata.srid
            lyr.save()


class Migration(migrations.Migration):

    dependencies = [
        ('raster', '0016_rasterlayermetadata_srid'),
    ]

    operations = [
        migrations.RunPython(move_srid_from_layer_to_metadata_forward, move_srid_from_layer_to_metadata_backward),
    ]



from django.conf import settings
from raster.models import RasterTile
from raster.tiles.const import WEB_MERCATOR_TILESIZE
from raster.tiles.utils import tile_bounds, tile_scale


def get_raster_tile(layer_id, tilez, tilex, tiley):
    """
    Get the raster from a tile for further processing. If the requested tile
    does not exists in the database, higher level tiles are searched. If a
    higher level tile is found, it is warped to the requested zoom level. This
    ensures that a tile can be requested at any zoom level.
    """
    # Loop through zoom levels to search for a tile
    for zoom in range(tilez, -1, -1):
        # Compute multiplier to find parent raster
        multiplier = 2 ** (tilez - zoom)
        # Fetch tile
        tile = RasterTile.objects.filter(
            tilex=tilex / multiplier,
            tiley=tiley / multiplier,
            tilez=zoom,
            rasterlayer_id=layer_id,
        )

        if tile.exists():
            # Extract raster from tile model
            result = tile[0].rast
            # If the tile is a parent of the original, warp it to the
            # original request tile.
            if zoom < tilez:
                # Compute bounds, scale and size of child tile
                bounds = tile_bounds(tilex, tiley, tilez)
                tilesize = int(getattr(settings, 'RASTER_TILESIZE', WEB_MERCATOR_TILESIZE))
                tilescale = tile_scale(tilez)

                # Warp parent tile to child tile in memory.
                result = result.warp({
                    'driver': 'MEM',
                    'width': tilesize,
                    'height': tilesize,
                    'scale': [tilescale, -tilescale],
                    'origin': [bounds[0], bounds[3]],
                })

            return result



from raster.utils import colormap_to_rgba


def set_session_colormap(session, key, colormap):
    """
    Store the colormap in the user session.
    """
    raster_legends = session.get('raster_legends', {})
    raster_legends[key] = {
        "colormap": colormap_to_rgba(colormap),
    }

    session['raster_legends'] = raster_legends


def get_session_colormap(session, key):
    """
    Get the colormap form a legend stored in the user session and identified
    by key.
    """
    raster_legends = session.get('raster_legends', {})
    legend = raster_legends.get(key)
    if legend:
        return legend['colormap']



# -*- coding: utf-8 -*-
import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('raster', '0029_auto_20160111_0646'),
    ]

    operations = [
        migrations.AlterField(
            model_name='rasterlayerbandmetadata',
            name='hist_values',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.BigIntegerField(), size=100),
        ),
    ]



# Generated by Django 2.1.7 on 2019-03-13 07:28

import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('raster', '0038_auto_20171116_1027'),
    ]

    operations = [
        migrations.AlterField(
            model_name='rasterlayerparsestatus',
            name='tile_levels',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.PositiveIntegerField(), default=list, size=None),
        ),
    ]



# -*- coding: utf-8 -*-
import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('raster', '0027_rasterlayerreprojected'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='rasterlayerparsestatus',
            name='tile_level',
        ),
        migrations.AddField(
            model_name='rasterlayerparsestatus',
            name='tile_levels',
            field=django.contrib.postgres.fields.ArrayField(default=[], base_field=models.PositiveIntegerField(), size=None),
        ),
    ]



import io
import json
import os
import re
import uuid
import zipfile
from datetime import datetime
from tempfile import NamedTemporaryFile

import numpy
from PIL import Image

from django.conf import settings
from django.contrib.gis.gdal import GDALRaster
from django.contrib.gis.gdal.raster.const import VSI_FILESYSTEM_BASE_PATH
from django.contrib.gis.geos import Polygon
from django.db.models import Max, Q
from django.http import FileResponse, Http404, HttpResponse
from django.shortcuts import get_object_or_404
from django.template.defaultfilters import slugify
from django.utils.functional import cached_property
from django.views.generic import View
from raster.algebra.const import ALGEBRA_PIXEL_TYPE_GDAL, BAND_INDEX_SEPARATOR
from raster.algebra.parser import RasterAlgebraParser
from raster.const import EXPORT_MAX_PIXELS, IMG_ENHANCEMENTS, IMG_FORMATS, MAX_EXPORT_NAME_LENGTH, README_TEMPLATE
from raster.exceptions import RasterAlgebraException
from raster.models import Legend, RasterLayer, RasterLayerBandMetadata, RasterLayerMetadata
from raster.shortcuts import get_session_colormap
from raster.tiles.const import WEB_MERCATOR_SRID, WEB_MERCATOR_TILESIZE
from raster.tiles.lookup import get_raster_tile
from raster.tiles.utils import tile_bounds, tile_index_range, tile_scale
from raster.utils import band_data_to_image, colormap_to_rgba, pixel_value_from_point


class RasterView(View):

    @property
    def is_pixel_request(self):
        return 'xcoord' in self.kwargs

    def get_colormap(self, layer=None):
        """
        Returns colormap from request and layer, looking for a colormap in the
        request or session, a custom legend name to construct the legend or the
        default colormap from the layer legend.
        """
        colormap = None

        if 'colormap' in self.request.GET:
            colormap = colormap_to_rgba(json.loads(self.request.GET['colormap']))
            # Ensure colormap range is in float format.
            if 'range' in colormap:
                colormap['range'] = (float(colormap['range'][0]), float(colormap['range'][1]))
        elif 'legend' in self.request.GET:
            store = self.request.GET.get('store', 'database')
            if store == 'session':
                colormap = get_session_colormap(
                    self.request.session,
                    self.request.GET['legend']
                )
            else:
                legend_input = self.request.GET['legend']
                try:
                    legend_input = int(legend_input)
                except ValueError:
                    pass

                # Try to get legend by id, name or from input layer
                if isinstance(legend_input, int):
                    legend = get_object_or_404(Legend, id=legend_input)
                else:
                    legend = Legend.objects.filter(title__iexact=legend_input).first()
                colormap = legend.colormap

        elif 'layer' in self.kwargs:
            # Get legend for the input layer.
            legend = Legend.objects.filter(rasterlayer=self.kwargs.get('layer')).first()

            if legend and hasattr(legend, 'colormap'):
                colormap = legend.colormap

        if not colormap:
            # Use a continous grayscale color scheme.
            colormap = {
                'continuous': True,
                'from': (0, 0, 0),
                'to': (255, 255, 255),
            }
        else:
            # Add layer level value range to continuous colormaps if it was
            # not provided manually.
            if 'continuous' in colormap and 'range' not in colormap:
                meta = RasterLayerBandMetadata.objects.filter(rasterlayer_id=self.kwargs.get('layer')).first()
                if meta:
                    colormap['range'] = (meta.min, meta.max)

            # Filter by custom entries if requested
            if colormap and 'entries' in self.request.GET:
                entries = self.request.GET['entries'].split(',')
                colormap = {k: v for (k, v) in colormap.items() if str(k) in entries}

        return colormap

    def get_format(self):
        """
        Returns image format requested.
        """
        return IMG_FORMATS[self.kwargs.get('frmt')]

    def enhance(self, img):
        for key, enhancer in IMG_ENHANCEMENTS.items():
            if key in self.request.GET:
                img = enhancer(img).enhance(float(self.request.GET.get(key)))
        return img

    def write_img_to_response(self, img, stats):
        """
        Writes rgba numpy array to http response.
        """
        # Get requested format and corresponding content type.
        frmt, content_type = self.get_format()
        # Enhance image if requested.
        img = self.enhance(img)
        # Save image to io buffer.
        with io.BytesIO() as output:
            img.save(output, format=frmt)
            # Create response with image content.
            response = HttpResponse(
                output.getvalue(),
                content_type=content_type,
            )
            # Add aggregation statistics to response headers.
            response['aggregation'] = json.dumps(stats)
            return response

    def get_tile(self, layer_id, zlevel=None):
        """
        Returns a tile for rendering. If the tile does not exists, higher
        level tiles are searched and warped to lower level if found.
        """
        if self.is_pixel_request:
            tilez = self.max_zoom
            # Derive the tile index from the input coordinates.
            xcoord = float(self.kwargs.get('xcoord'))
            ycoord = float(self.kwargs.get('ycoord'))
            bbox = [xcoord, ycoord, xcoord, ycoord]
            indexrange = tile_index_range(bbox, tilez)
            tilex = indexrange[0]
            tiley = indexrange[1]
        else:
            # Get tile indices from the request url parameters.
            tilez = int(self.kwargs.get('z'))
            tilex = int(self.kwargs.get('x'))
            tiley = int(self.kwargs.get('y'))

        return get_raster_tile(layer_id, tilez, tilex, tiley)

    def get_layer(self):
        """
        Gets layer from request data trying both name and id.
        """
        # Get layer query data from input
        if 'layer' in self.kwargs:
            data = self.kwargs.get('layer')
        elif 'layer' in self.request.GET:
            data = self.request.GET.get('layer')
        else:
            raise Http404
        # Determine query paremeter type
        try:
            data = int(data)
            query = Q(id=data)
        except ValueError:
            query = Q(rasterfile__contains='rasters/' + data)

        return get_object_or_404(RasterLayer, query)

    @cached_property
    def max_zoom(self):
        return RasterLayerMetadata.objects.filter(
            rasterlayer_id__in=self.get_ids().values()
        ).aggregate(zlevel=Max('max_zoom'))['zlevel']


class AlgebraView(RasterView):
    """
    A view to calculate map algebra on raster layers.
    """

    _layer_ids = None

    def get_ids(self):
        if self._layer_ids is not None:
            return self._layer_ids

        if 'layer' in self.kwargs:
            # For tms requests, construct simple ids dictionary.
            data = self.kwargs.get('layer')
            # Determine query paremeter type
            try:
                layer_id = int(data)
            except ValueError:
                query = Q(rasterfile__contains='rasters/' + data)
                layer_id = get_object_or_404(RasterLayer, query).id
            # For TMS tile request, get the layer id from the url.
            self._layer_ids = {'x': layer_id}
        else:
            # For algebra requests, get the layer ids from the query parameter.
            ids = self.request.GET.get('layers', '').split(',')

            # Check if layer parameter is valid
            if not len(ids) or not all('=' in idx for idx in ids):
                raise RasterAlgebraException('Layer parameter is not valid.')

            # Split id/name input pairs
            ids = [idx.split('=') for idx in ids]

            # Convert ids to integer
            try:
                ids = {idx[0]: int(idx[1]) for idx in ids}
            except ValueError:
                raise RasterAlgebraException('Layer parameter is not valid.')
            self._layer_ids = ids

        return self._layer_ids

    def get_formula(self):
        if 'layer' in self.kwargs:
            # Set the formula to trivial for TMS requests.
            return 'x'
        else:
            return self.request.GET.get('formula', None)

    def get(self, request, *args, **kwargs):
        # Get layer ids
        ids = self.get_ids()

        # Prepare unique list of layer ids to be efficient if the same layer
        # is used multiple times (for band access for instance).
        layerids = set(ids.values())

        # Get the tiles for each unique layer.
        tiles = {}
        for layerid in layerids:
            tile = self.get_tile(layerid)
            if tile:
                tiles[layerid] = tile
            else:
                # Create empty image if any layer misses the required tile
                img = Image.new("RGBA", (WEB_MERCATOR_TILESIZE, WEB_MERCATOR_TILESIZE), (0, 0, 0, 0))
                return self.write_img_to_response(img, {})

        # Map tiles to a dict with formula names as keys.
        data = {}
        for name, layerid in ids.items():
            data[name] = tiles[layerid]

        # Get formula from request
        formula = self.get_formula()

        # Dispatch by request type. If a formula was provided, use raster
        # algebra otherwise look for rgb request.
        if formula:
            return self.get_algebra(data, formula)
        else:
            keys = [key.split(BAND_INDEX_SEPARATOR)[0] for key in data.keys()]
            if 'r' in keys and 'g' in keys and 'b' in keys:
                return self.get_rgb(data)
            else:
                raise RasterAlgebraException(
                    'Specify raster algebra formula or provide rgb layer keys.'
                )

    def get_algebra(self, data, formula):
        parser = RasterAlgebraParser()

        # Evaluate raster algebra expression, return 400 if not successful
        try:
            # Evaluate raster algebra expression
            result = parser.evaluate_raster_algebra(data, formula)
        except:
            raise RasterAlgebraException('Failed to evaluate raster algebra.')

        # For pixel value requests, return result as json.
        if self.is_pixel_request:
            xcoord = float(self.kwargs.get('xcoord'))
            ycoord = float(self.kwargs.get('ycoord'))
            val = pixel_value_from_point(result, [xcoord, ycoord])
            return HttpResponse(
                json.dumps({'x': xcoord, 'y': ycoord, 'value': val}),
                content_type='application/json',
            )

        # For tif requests, skip colormap and return georeferenced tif file.
        if self.kwargs.get('frmt') == 'tif':
            vsi_path = os.path.join(VSI_FILESYSTEM_BASE_PATH, str(uuid.uuid4()))
            rast = result.warp({
                'name': vsi_path,
                'driver': 'tif',
                'compress': 'DEFLATE',
            })
            content_type = IMG_FORMATS['tif'][1]
            return HttpResponse(rast.vsi_buffer, content_type)

        # Get array from algebra result
        if result.bands[0].nodata_value is None:
            result = result.bands[0].data()
        else:
            result = numpy.ma.masked_values(
                result.bands[0].data(),
                result.bands[0].nodata_value,
            )

        # Get colormap.
        colormap = self.get_colormap()

        # Render tile using the legend data
        img, stats = band_data_to_image(result, colormap)

        # Return rendered image
        return self.write_img_to_response(img, stats)

    def get_rgb_scale(self):
        if 'scale' not in self.request.GET:
            return

        # The scale is either a number or two numbers separated by comma.
        scale = self.request.GET.get('scale').split(',')
        if len(scale) == 1:
            scale_min = 0
            scale_max = float(scale[0])
        else:
            # Get min an max scale from
            scale_min = float(scale[0])
            scale_max = float(scale[1])

        return scale_min, scale_max

    def get_alpha(self):
        return 'alpha' in self.request.GET

    def get_rgb(self, data):
        # Get data arrays from tiles, by band if requested.
        for key, tile in data.items():

            keysplit = key.split(BAND_INDEX_SEPARATOR)
            variable = keysplit[0]

            if len(keysplit) > 1:
                band_index = int(keysplit[1])
            else:
                band_index = 0

            band = tile.bands[band_index]
            if variable == 'r':
                red = band.data()
                red_nodata = band.nodata_value
                red_gdal_dtype = band.datatype()
            elif variable == 'g':
                green = band.data()
                green_nodata = band.nodata_value
            elif variable == 'b':
                blue = band.data()
                blue_nodata = band.nodata_value

        # For tif requests, skip rgb rendering and return georeferenced tif file.
        if self.kwargs.get('frmt') == 'tif':
            vsi_path = os.path.join(VSI_FILESYSTEM_BASE_PATH, str(uuid.uuid4()))
            # Construct 3 band raster, assuming all
            ref = next(iter(data.values()))
            result = GDALRaster({
                'name': vsi_path,
                'driver': 'tif',
                'srid': WEB_MERCATOR_SRID,
                'datatype': red_gdal_dtype,
                'width': ref.width,
                'height': ref.height,
                'origin': ref.origin,
                'scale': ref.scale,
                'skew': ref.skew,
                'bands': [
                    {'data': red, 'nodata_value': red_nodata},
                    {'data': green.astype(red.dtype), 'nodata_value': green_nodata},
                    {'data': blue.astype(red.dtype), 'nodata_value': blue_nodata},
                ],
                'papsz_options': {
                    'compress': 'deflate',
                },
            })
            content_type = IMG_FORMATS['tif'][1]
            return HttpResponse(result.vsi_buffer, content_type)

        # Get scale for the image value range.
        scale = self.get_rgb_scale()
        if scale is not None:
            # Clip the image minimum.
            red[red < scale[0]] = scale[0]
            green[green < scale[0]] = scale[0]
            blue[blue < scale[0]] = scale[0]

            # Clip the image maximum.
            red[red > scale[1]] = scale[1]
            green[green > scale[1]] = scale[1]
            blue[blue > scale[1]] = scale[1]

            # Scale the image.
            red = 255 * (red - scale[0]) / scale[1]
            green = 255 * (green - scale[0]) / scale[1]
            blue = 255 * (blue - scale[0]) / scale[1]

        if self.get_alpha():
            mode = 'RGBA'
            reshape = 4
            # Create the alpha channel.
            alpha = 255 * (red > 0) * (blue > 0) * (green > 0)
            img_array = numpy.array((red.ravel(), green.ravel(), blue.ravel(), alpha.ravel()))
        else:
            mode = 'RGB'
            reshape = 3
            img_array = numpy.array((red.ravel(), green.ravel(), blue.ravel()))

        # Reshape array into tile size.
        img_array = img_array.T.reshape(WEB_MERCATOR_TILESIZE, WEB_MERCATOR_TILESIZE, reshape).astype('uint8')

        # Create image from array
        img = Image.fromarray(img_array, mode=mode)
        stats = {}

        # Return rendered image
        return self.write_img_to_response(img, stats)


class LegendView(RasterView):

    def get(self, request, legend_id):
        """
        Returns the legend for this layer as a json string. The legend is a list of
        legend entries with the attributes "name", "expression" and "color".
        """
        if legend_id:
            # Get legend from id
            legend = get_object_or_404(Legend, id=legend_id)
        else:
            # Try to get legend from layer
            lyr = self.get_layer()
            if not lyr.legend:
                raise Http404
            legend = lyr.legend

        return HttpResponse(legend.json, content_type='application/json')


class ExportView(AlgebraView):

    def construct_raster(self, z, xmin, xmax, ymin, ymax):
        """
        Create an empty tif raster file on disk using the input tile range. The
        new raster aligns with the xyz tile scheme and can be filled
        sequentially with raster algebra results.
        """
        # Compute bounds and scale to construct raster.
        bounds = []
        for x in range(xmin, xmax + 1):
            for y in range(ymin, ymax + 1):
                bounds.append(tile_bounds(x, y, z))
        bounds = [
            min([bnd[0] for bnd in bounds]),
            min([bnd[1] for bnd in bounds]),
            max([bnd[2] for bnd in bounds]),
            max([bnd[3] for bnd in bounds]),
        ]
        scale = tile_scale(z)
        # Create tempfile.
        raster_workdir = getattr(settings, 'RASTER_WORKDIR', None)
        self.exportfile = NamedTemporaryFile(dir=raster_workdir, suffix='.tif')
        # Instantiate raster using the tempfile path.
        return GDALRaster({
            'srid': WEB_MERCATOR_SRID,
            'width': (xmax - xmin + 1) * WEB_MERCATOR_TILESIZE,
            'height': (ymax - ymin + 1) * WEB_MERCATOR_TILESIZE,
            'scale': (scale, -scale),
            'origin': (bounds[0], bounds[3]),
            'driver': 'tif',
            'bands': [{'data': [0], 'nodata_value': 0}],
            'name': self.exportfile.name,
            'datatype': ALGEBRA_PIXEL_TYPE_GDAL,
        })

    def get_tile_range(self):
        """
        Compute a xyz tile range from the query parameters. If no bbox
        parameter is found, the range defaults to the maximum extent of
        all input raster layers.
        """
        # Get raster layers
        layers = RasterLayer.objects.filter(id__in=self.get_ids().values())
        # Establish zoom level
        if self.request.GET.get('zoom', None):
            zlevel = int(self.request.GET.get('zoom'))
        else:
            # Get highest zoom level of all input layers
            zlevel = self.max_zoom
        # Use bounding box to compute tile range
        if self.request.GET.get('bbox', None):
            bbox = Polygon.from_bbox(self.request.GET.get('bbox').split(','))
            bbox.srid = 4326
            bbox.transform(WEB_MERCATOR_SRID)
            tile_range = tile_index_range(bbox.extent, zlevel)
        else:
            # Get list of tile ranges
            layer_ranges = []
            for layer in layers:
                layer_ranges.append(tile_index_range(layer.extent(), zlevel))
            # Estabish overlap of tile index ranges
            tile_range = [
                min([rng[0] for rng in layer_ranges]),
                min([rng[1] for rng in layer_ranges]),
                max([rng[2] for rng in layer_ranges]),
                max([rng[3] for rng in layer_ranges]),
            ]
        return [zlevel, ] + tile_range

    def write_colormap(self, zfile):
        # Try to get colormap
        colormap = self.get_colormap()
        # Set a simple header for this colormap
        colorstr = '# Raster Algebra Colormap\n'
        # Check if this is a continuous legend.
        colorstr += 'INTERPOLATION: ' + ('CONTINUOUS' if colormap.pop('continuous', None) else 'DISCRETE') + '\n'
        # Add expressions and colors of the colormap
        for key, val in colormap.items():
            colorstr += str(key) + ',' + ','.join((str(x) for x in val)) + ',' + str(key) + '\n'
        # Write colormap file
        zfile.writestr('COLORMAP.txt', colorstr)

    def write_readme(self, zfile):
        # Get tile index range
        zoom, xmin, ymin, xmax, ymax = self.get_tile_range()
        # Construct layer names string
        layerstr = ''
        for name, layerid in self.get_ids().items():
            layer = RasterLayer.objects.get(id=layerid)
            layerstr += '{layerid} "{name}" (Formula label: {label})\n'.format(
                name=layer.name,
                label=name,
                layerid=layerid
            )
        # Get description, append newline if provided
        description = self.request.GET.get('description', '')
        if description:
            description += '\n'
        # Initiate metadata object
        readmedata = {
            'datetime': datetime.now().strftime('%Y-%m-%d at %H:%M'),
            'url': self.request.build_absolute_uri(),
            'bbox': self.request.GET.get('bbox', 'Minimum bounding-box covering all layers.'),
            'formula': self.request.GET.get('formula'),
            'zoom': str(zoom),
            'xindexrange': '{} - {}'.format(xmin, xmax),
            'yindexrange': '{} - {}'.format(ymin, ymax),
            'layers': layerstr,
            'description': description,
        }
        # Write readme file
        readme = README_TEMPLATE.format(**readmedata)
        zfile.writestr('README.txt', readme)

    def get(self, request):
        # Initiate algebra parser
        parser = RasterAlgebraParser()
        # Get formula from request
        formula = request.GET.get('formula')
        # Get id list from request
        ids = self.get_ids()
        # Compute tile index range
        zoom, xmin, ymin, xmax, ymax = self.get_tile_range()
        # Check maximum size of target raster in pixels
        max_pixels = getattr(settings, 'RASTER_EXPORT_MAX_PIXELS', EXPORT_MAX_PIXELS)
        if WEB_MERCATOR_TILESIZE * (xmax - xmin) * WEB_MERCATOR_TILESIZE * (ymax - ymin) > max_pixels:
            raise RasterAlgebraException('Export raster too large.')
        # Construct an empty raster with the output dimensions
        result_raster = self.construct_raster(zoom, xmin, xmax, ymin, ymax)
        target = result_raster.bands[0]
        # Get raster data as 1D arrays and store in dict that can be used
        # for formula evaluation.
        for xindex, x in enumerate(range(xmin, xmax + 1)):
            for yindex, y in enumerate(range(ymin, ymax + 1)):
                data = {}
                for name, layerid in ids.items():
                    tile = get_raster_tile(layerid, zoom, x, y)
                    if tile:
                        data[name] = tile
                # Ignore this tile if data is not found for all layers
                if len(data) != len(ids):
                    continue
                # Evaluate raster algebra expression, return 400 if not successful
                try:
                    # Evaluate raster algebra expression
                    tile_result = parser.evaluate_raster_algebra(data, formula)
                except:
                    raise RasterAlgebraException('Failed to evaluate raster algebra.')
                # Update nodata value on target
                target.nodata_value = tile_result.bands[0].nodata_value
                # Update results raster with algebra
                target.data(
                    data=tile_result.bands[0].data(),
                    size=(WEB_MERCATOR_TILESIZE, WEB_MERCATOR_TILESIZE),
                    offset=(xindex * WEB_MERCATOR_TILESIZE, yindex * WEB_MERCATOR_TILESIZE),
                )
        # Create filename base with datetime stamp
        filename_base = 'algebra_export'
        # Add name slug to filename if provided
        if request.GET.get('filename', ''):
            # Sluggify name
            slug = slugify(request.GET.get('filename'))
            # Remove all unwanted characters
            slug = "".join([c for c in slug if re.match(r'\w|\-', c)])
            # Limit length of custom name slug
            slug = slug[:MAX_EXPORT_NAME_LENGTH]
            # Add name slug to filename base
            filename_base += '_' + slug
        filename_base += '_{0}'.format(datetime.now().strftime('%Y_%m_%d_%H_%M'))
        # Compress resulting raster file into a zip archive
        raster_workdir = getattr(settings, 'RASTER_WORKDIR', None)
        dest = NamedTemporaryFile(dir=raster_workdir, suffix='.zip')
        dest_zip = zipfile.ZipFile(dest.name, 'w', allowZip64=True)
        dest_zip.write(
            filename=self.exportfile.name,
            arcname=filename_base + '.tif',
            compress_type=zipfile.ZIP_DEFLATED,
        )
        # Write README.txt and COLORMAP.txt files to zip file
        self.write_readme(dest_zip)
        self.write_colormap(dest_zip)
        # Close zip file before returning
        dest_zip.close()
        # Create file based response containing zip file and return for download
        response = FileResponse(
            open(dest.name, 'rb'),
            content_type='application/zip'
        )
        response['Content-Disposition'] = 'attachment; filename="{0}"'.format(filename_base + '.zip')
        return response



# -*- coding: utf-8 -*-
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('raster', '0025_auto_20151113_0259'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='rasterlayerbandmetadata',
            unique_together=set([('rasterlayer', 'band')]),
        ),
    ]



import shutil
import traceback

from celery import group, task

from django.conf import settings
from raster.tiles.const import GLOBAL_MAX_ZOOM_LEVEL, MIN_ZOOMLEVEL_TASK_PARALLEL
from raster.tiles.parser import RasterLayerParser


@task
def create_tiles(rasterlayer_id, zoom, extract_metadata=False):
    """
    Create all tiles for a raster layer at the input zoom level.
    """
    try:
        parser = RasterLayerParser(rasterlayer_id)

        # Open raster file and extract metadata if requested.
        if extract_metadata:
            parser.open_raster_file()
            parser.extract_metadata()

        # Check if zoom level should be built.
        if zoom is None:
            zoom = parser.max_zoom
        elif isinstance(zoom, (int, float)):
            if zoom > parser.max_zoom:
                return
        else:
            zoom = [zl for zl in zoom if zl <= parser.max_zoom]
            if not len(zoom):
                return
        # Open raster file if not open already.
        if not extract_metadata:
            parser.open_raster_file()

        parser.reproject_rasterfile()
        parser.create_tiles(zoom)
    except:
        parser.log(
            traceback.format_exc(),
            status=parser.rasterlayer.parsestatus.FAILED
        )
        raise
    finally:
        if hasattr(parser, 'tmpdir'):
            shutil.rmtree(parser.tmpdir)


@task
def clear_tiles(rasterlayer_id):
    """
    Drop all tiles of a rasterlayer.
    """
    parser = RasterLayerParser(rasterlayer_id)
    parser.drop_all_tiles()


@task
def send_success_signal(rasterlayer_id):
    """
    Drop empty tiles of a raster layer and send parse succes signal.
    """
    parser = RasterLayerParser(rasterlayer_id)
    parser.send_success_signal()


@task
def all_in_one(rasterlayer_id, zoom_range):
    """
    Parses raster in a single task.
    """
    clear_tiles(rasterlayer_id)
    create_tiles(rasterlayer_id, zoom_range, True)
    send_success_signal(rasterlayer_id)


def parse(rasterlayer_id):
    """
    Parse raster layer to extract metadata and create tiles.
    """
    parser = RasterLayerParser(rasterlayer_id)
    parser.log('Started parsing raster.')

    # Create array of all allowed zoom levels
    if parser.rasterlayer.build_pyramid:
        zoom_range = list(range(GLOBAL_MAX_ZOOM_LEVEL + 1))
    else:
        if parser.rasterlayer.max_zoom is not None:
            zoom_range = (parser.rasterlayer.max_zoom, )
        else:
            zoom_range = None

    # Check if parsing should happen asynchronously
    parse_async = getattr(settings, 'RASTER_USE_CELERY', False)
    parse_single_task = getattr(settings, 'RASTER_PARSE_SINGLE_TASK', False)
    if parse_async and not parse_single_task:
        if zoom_range is not None:
            # Bundle the first five raster layers to one task. For low zoom
            # levels, downloading is more costly than parsing.
            create_tiles_chain = create_tiles.si(rasterlayer_id, zoom_range[:MIN_ZOOMLEVEL_TASK_PARALLEL], True)

            # Run the higher level zooms in parallel tasks through a task group.
            if len(zoom_range) > MIN_ZOOMLEVEL_TASK_PARALLEL:
                high_zoom_level_group = group(
                    create_tiles.si(rasterlayer_id, zoom) for zoom in zoom_range[MIN_ZOOMLEVEL_TASK_PARALLEL:]
                )
                # Combine bundle and middle levels to priority group.
                create_tiles_chain = (create_tiles_chain | high_zoom_level_group)

        else:
            create_tiles_chain = create_tiles.si(rasterlayer_id, None, True)

        # Setup the parser logic as parsing chain
        parsing_task_chain = (
            clear_tiles.si(rasterlayer_id)
            | create_tiles_chain
            | send_success_signal.si(rasterlayer_id)
        )

        # Apply the parsing chain
        parser.log('Parse task queued, waiting for worker availability.')
        parsing_task_chain.apply_async()
    elif parse_async and parse_single_task:
        parser.log('Parse task queued in all-in-one mode, waiting for worker availability.')
        all_in_one.delay(rasterlayer_id, zoom_range)
    else:
        all_in_one(rasterlayer_id, zoom_range)



# -*- coding: utf-8 -*-
# Generated by Django 1.9b1 on 2015-11-10 08:25
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('raster', '0022_auto_20151110_0810'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='rasterlayer',
            name='parse_log',
        ),
    ]



# -*- coding: utf-8 -*-
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('raster', '0017_copy_srids_20150922_0207'),
    ]

    operations = [
        migrations.AddField(
            model_name='rasterlayermetadata',
            name='srs_wkt',
            field=models.TextField(null=True, blank=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2017-11-16 10:27
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('raster', '0037_auto_20170509_0559'),
    ]

    operations = [
        migrations.AlterField(
            model_name='rasterlayer',
            name='source_url',
            field=models.CharField(blank=True, default='', help_text='External url to get the raster file from. If a value is set,the rasterfile field will be ignored.', max_length=2500),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.8.dev20160616122412 on 2016-07-01 04:22
import json
from operator import itemgetter

import django.db.models.deletion
from django.db import migrations, models


def move_legend_entries_to_through_model(apps, schema_editor):
    """
    Move the data from the normal m2m table to the through table for legend entries.
    """
    Legend = apps.get_model("raster", "Legend")
    LegendEntryOrder = apps.get_model("raster", "LegendEntryOrder")
    for leg in Legend.objects.all():
        for entry in leg.entries.all():
            # Respect unique contraint.
            if LegendEntryOrder.objects.filter(legend=leg, legendentry=entry).exists():
                continue
            LegendEntryOrder.objects.create(
                legend=leg,
                legendentry=entry,
                code='',
            )
        # Make sure legend json is up to date.
        data = []
        for val in leg.legendentryorder_set.prefetch_related('legendentry').all():
            data.append({
                'name': val.legendentry.semantics.name,
                'expression': val.legendentry.expression,
                'code': val.code,
                'color': val.legendentry.color,
            })
        data = sorted(data, key=itemgetter('name'))
        leg.json = json.dumps(data)
        leg.save()


def move_legend_entries_from_through_model_to_normal_m2m(apps, schema_editor):
    """
    Move the data from the entries order table to the through table for legend entries.
    """
    Legend = apps.get_model("raster", "Legend")
    LegendEntryOrder = apps.get_model("raster", "LegendEntryOrder")
    for leg in Legend.objects.all():
        for enor in LegendEntryOrder.objects.filter(legend=leg):
            leg.entries.add(enor.legendentry)


class Migration(migrations.Migration):

    dependencies = [
        ('raster', '0033_auto_20160510_1041'),
    ]

    operations = [
        migrations.CreateModel(
            name='LegendEntryOrder',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.CharField(max_length=100, default='', blank=True)),
                ('legendentry', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='raster.LegendEntry')),
                ('legend', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='raster.Legend')),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='legendentryorder',
            unique_together=set([('legend', 'legendentry')]),
        ),
        migrations.RunPython(move_legend_entries_to_through_model, move_legend_entries_from_through_model_to_normal_m2m),
        migrations.RemoveField(
            model_name='legend',
            name='entries',
        ),
        migrations.AddField(
            model_name='legend',
            name='entries',
            field=models.ManyToManyField(through='raster.LegendEntryOrder', to='raster.LegendEntry'),
        ),
    ]



# -*- coding: utf-8 -*-
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('raster', '0024_auto_20151111_0532'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='rastertile',
            name='filename',
        ),
        migrations.AlterField(
            model_name='legendentry',
            name='expression',
            field=models.CharField(help_text=b'Use a number or a valid numpy logical expression where x is thepixel value. For instance: "(-3.0 < x) & (x <= 1)" or "x <= 1".', max_length=500),
        ),
        migrations.AlterField(
            model_name='rasterlayer',
            name='nodata',
            field=models.CharField(help_text=b'Leave blank to keep the internal band nodata values. If a nodatavalue is specified here, it will be used for all bands of this raster.', max_length=100, null=True, blank=True),
        ),
    ]



# -*- coding: utf-8 -*-
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('raster', '0009_rasterlayer_max_zoom'),
    ]

    operations = [
        migrations.AlterField(
            model_name='legendentry',
            name='expression',
            field=models.CharField(help_text=b'Use a number or a valid numpy logical expression where x is the pixel value. For instance: "(-3.0 < x) & (x <= 1)" or "x <= 1".', max_length=500),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='rasterlayer',
            name='rasterfile',
            field=models.FileField(null=True, upload_to=b'rasters', blank=True),
            preserve_default=True,
        ),
    ]
